MODULE basis
  USE atomdata
  USE calcpotential
  USE Globalmath
  USE pseudo
  USE wavesolve

  IMPLICIT NONE

CONTAINS

  SUBROUTINE setbasis(Grid,Pot,Orbit,FC,PAW)
    TYPE(GridInfo), INTENT(IN) :: Grid
    TYPE(PotentialInfo), INTENT(IN) :: Pot
    TYPE(OrbitInfo), INTENT(IN) :: Orbit
    TYPE(FCInfo), INTENT(IN) :: FC
    TYPE(PseudoInfo), INTENT(INOUT) :: PAW


    INTEGER :: n,irc,norbit,nbase,l,lmax,mxbase
    INTEGER :: i,j,k,io,ok,nbl,nr,nodes,ib,loop,niter,iter
    REAL(8) :: h,rc,q00,energy,rat,delta,thisconv,qeff,tq
    REAL(8) :: ecoul,etxc,eexc
    CHARACTER(1) :: answer
    REAL(8), POINTER  :: r(:)

    n=Grid%n
    h=Grid%h
    r=>Grid%r
    irc=PAW%irc
    nr=irc+20
    rc=PAW%rc
    lmax=PAW%lmax

    norbit=Orbit%norbit
    nbase=0

    DO l=0,lmax
       DO io=1,norbit    ! cycle through all configuration
          IF (Orbit%l(io).EQ.l.AND.FC%iscore(io).EQ.1) THEN
             nbase=nbase+1
          ENDIF
       ENDDO
    ENDDO

    mxbase=nbase+5*PAW%lmax
    WRITE(6,*) 'Found ', nbase,' valence basis functions '
    WRITE(6,*) 'Allocating for ', mxbase, ' total basis functions'
    ALLOCATE(PAW%phi(n,mxbase),PAW%tphi(n,mxbase),PAW%tp(n,mxbase),&
         PAW%np(mxbase),PAW%l(mxbase),PAW%eig(mxbase),PAW%occ(mxbase),&
         PAW%ck(mxbase),stat=ok)
    IF (ok /= 0) THEN
       WRITE(6,*) 'Error in  PAW%phi allocation', n,mxbase,ok
       STOP
    ENDIF
    !
    PAW%phi=0; PAW%tphi=0; PAW%tp=0;  PAW%ck=0
    PAW%np=0;  PAW%l=0; PAW%eig=0;  PAW%occ=0
    nbase=0
    WRITE(6,*) '  basis functions:'
    WRITE(6,*)' No.   n     l         energy         occ   '
    DO l=0,lmax
       nbl=0
       DO io=1,norbit    ! cycle through all configuration
          IF (Orbit%l(io).EQ.l.AND.FC%iscore(io).EQ.1) THEN
             nbl=nbl+1
             nbase=nbase+1
             PAW%np(nbase)=Orbit%np(io)
             PAW%l(nbase)=l
             PAW%eig(nbase)=Orbit%eig(io)
             PAW%occ(nbase)=Orbit%occ(io)
             PAW%phi(:,nbase)=Orbit%wfn(:,io)
             WRITE(6,'(3i6,1p2e15.6)') nbase,PAW%np(nbase),l,             &
                  PAW%eig(nbase),PAW%occ(nbase)
          ENDIF
       ENDDO
       generalizedloop: DO
          WRITE(6,*) 'For l = ',l,' there are currently ',nbl,&
               'basis functions'
          WRITE(6,*) 'enter y to add additional functions or n to ' &
               ,'go to next l'
          READ(5,*) answer
          IF (answer.NE.'y') EXIT generalizedloop
          WRITE(6,*) 'enter energy for generalized function'
          READ(5,*) energy
          IF(energy.LT.0.d0) THEN
             WRITE(6,*) 'energy is negative',energy,' -- program terminated'
             STOP
          ENDIF
          nbase=nbase+1
          IF (nbase > mxbase ) THEN
             WRITE(6,*) 'Error in  setbasis -- too many functions ', nbase,mxbase
             STOP
          ENDIF
          PAW%l(nbase)=l
          PAW%np(nbase)=999
          PAW%eig(nbase)=energy
          PAW%occ(nbase)=0.d0
          PAW%phi(1:n,nbase)=0.d0
          CALL unboundsch(Grid,Pot,n,l,energy,PAW%phi(:,nbase),nodes)
          rat=MAX(ABS(PAW%phi(irc,nbase)),ABS(PAW%phi(irc+1,nbase)))
          rat=DSIGN(rat,PAW%phi(irc,nbase))
          PAW%phi(1:n,nbase)=PAW%phi(1:n,nbase)/rat
          WRITE(6,'(3i6,1p2e15.6)') nbase,PAW%np(nbase),l,             &
               PAW%eig(nbase),PAW%occ(nbase)
          nbl=nbl+1
       ENDDO generalizedloop
       !
    ENDDO   ! end lmax loop

    WRITE(6,*) 'completed phi basis with ',nbase,' functions '

    PAW%nbase=nbase

    ALLOCATE(PAW%oij(nbase,nbase),PAW%dij(nbase,nbase),stat=i)
    IF (i/=0) THEN
       WRITE(6,*) 'Error in allocating oij,dij ', nbase,i
       STOP
    ENDIF

    PAW%oij=0; PAW%dij=0

  END SUBROUTINE setbasis

  !**************************************************************************
  !  Version using Blochl's form of projector and orthogonalization procedure
  !   At the end of this subroutine, the basis functions and projectors are
  !    orthogonalized
  !  program to generate atomic basis functions for Blochl's paw formalism
  !    starting with the self-consistent density functional
  !    atom ground state for atom with atomic number nz, ionic charge qion
  !    choose bound and continuum basis functions phi
  !    accumulate core and valence densities
  !**************************************************************************
  SUBROUTINE makebasis(Grid,PAW)
    TYPE(GridInfo), INTENT(IN) :: Grid
    TYPE(PseudoInfo), INTENT(INOUT) :: PAW

    INTEGER :: n,irc,nr
    INTEGER :: i,j,k,io,ok,lmax
    REAL(8) :: h,rc
    REAL(8), ALLOCATABLE :: denout(:),tmp(:),VNC(:)
    TYPE(PotentialInfo), TARGET:: PS
    REAL(8), POINTER  :: r(:)

    n=Grid%n
    h=Grid%h
    r=>Grid%r
    irc=PAW%irc
    nr=irc+20
    rc=PAW%rc
    lmax=PAW%lmax
    PS%nz=0

    ALLOCATE(tmp(n),VNC(n),PS%den(n),PS%rv(n),stat=ok)
    IF (ok /= 0 ) THEN
       WRITE(6,*) 'Error in allocating  denout in makebasis ',n,ok
       STOP
    ENDIF

    ! find basis functions
    VNC(1:n)=PAW%rveff(1:n)/r(1:n)
    CALL nderiv(h,VNC(1:15),tmp(1:15),15,ok)
    PS%v0=VNC(3)+3*(VNC(1)-VNC(2))
    PS%v0p=tmp(3)+3*(tmp(1)-tmp(2))
    PS%rv=PAW%rveff
    !write(6,*) 'VNC  v0 ', PS%v0,PS%v0p,PS%rv(5)
    CALL formprojectors(Grid,PS,PAW)

    DEALLOCATE(tmp,VNC,PS%den,PS%rv)

  END SUBROUTINE makebasis

  SUBROUTINE SCbasis(Grid,FC,nz,PAW)
    TYPE(GridInfo), INTENT(IN) :: Grid
    TYPE(FCInfo), INTENT(IN) :: FC
    INTEGER, INTENT(IN) :: nz
    TYPE(PseudoInfo), INTENT(INOUT) :: PAW

    INTEGER :: iv,niter,loop,i,n,nbase,io,index
    REAL(8) :: q00,tq,h,conv,ecoul,etxc,eexc,delta
    REAL(8), ALLOCATABLE :: denout(:),dum(:)
    TYPE(PotentialInfo) :: PS

    h=Grid%h
    n=Grid%n
    nbase=PAW%nbase

    DO
       WRITE(6,*) 'Input index of basis function for setting vloc'
         READ(5,*) index
         IF (index>=1.AND.index<=nbase) EXIT
      ENDDO

      WRITE(PAW%Vloc_description,&
           '("Fixed shape vloc with amplitude set for basis index",i4)')&
           index
      WRITE(6,*) PAW%Vloc_description

      ALLOCATE(PS%rv(n),PS%den(n),denout(n),dum(n),stat=i)
      IF (i/=0) THEN
         WRITE(6,*) 'Allocation error in SCbasis ', n,i
         STOP
      ENDIF

      PS%nz=0
      PAW%vloc=0
      DO io=1,nbase
         PAW%tphi(:,io)=PAW%phi(:,io)
      ENDDO

      PAW%den=0; PAW%tden=0
      DO io=1,nbase
         PAW%den(:)=PAW%den(:)+PAW%occ(io)*(PAW%phi(:,io))**2
         PAW%tden(:)=PAW%tden(:)+PAW%occ(io)*(PAW%tphi(:,io))**2
      ENDDO

      DO iv=1,2

         conv=nbase*1.d-11
         niter=100
         loop=0
         delta=1.d10
         denout=PAW%tden

         tphiloop: DO
            loop=loop+1
            IF(loop > niter) THEN
               WRITE(6,*) 'SCbasis pgm terminating in tphiloop ', loop
               STOP
            ENDIF


            tq=overint(n,h,PAW%tden)
            WRITE(6,*) 'tq = ',tq
            q00=-nz+FC%zcore+FC%zvale-tq
            WRITE(6,*) 'q00 = ',q00

            PS%q=tq
            PS%den=PAW%tden
            CALL potential(Grid,PS,ecoul,etxc,eexc)

            PAW%rveff(:)=PS%rv(:)+PAW%vloc(:)*Grid%r(:)+q00*PAW%hatpot(:)

            IF (loop>4.AND. delta<conv) THEN
               WRITE(6,*) 'Projectors converged with loop,delta = ', loop, delta
               EXIT tphiloop
            ENDIF

            CALL makebasis(Grid,PAW)

            PAW%tden=0
            DO io=1,nbase
               PAW%tden(:)=PAW%tden(:)+PAW%occ(io)*(PAW%tphi(:,io))**2
            ENDDO

            dum=denout-PAW%tden
            delta=SQRT(DOT_PRODUCT(dum(1:n),dum(1:n)))
            WRITE(6,*) 'loop = ', loop,' delta = ', delta
            denout=PAW%tden

         ENDDO tphiloop

         IF (iv==1) THEN
            WRITE(6,*) 'Completed first tphiloop', PAW%ck(1:nbase)
            WRITE(6,*) 'Resetting vloc with vlocfac == ', -PAW%ck(index)
            PAW%vloc=-PAW%ck(index)*PAW%projshape
         ENDIF

      ENDDO   !iv loop
      DEALLOCATE(PS%den,PS%rv,denout,dum)
    END SUBROUTINE SCbasis

    SUBROUTINE FindVlocfromVeff(Grid,FC,nz,PAW)
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(FCInfo), INTENT(IN) :: FC
      INTEGER, INTENT(IN) :: nz
      TYPE(PseudoInfo), INTENT(INOUT) :: PAW

      TYPE(PotentialInfo), TARGET:: PS
      REAL(8), POINTER  :: r(:)
      REAL(8) :: h,qeff,tq,rat,q00,ecoul,etxc,eexc
      INTEGER :: n,i,irc,io,nbase

      n=Grid%n
      h=Grid%h
      r=>Grid%r
      irc=PAW%irc
      PAW%den=0
      PAW%tden=0
      nbase=PAW%nbase
      DO io=1,nbase
         IF (PAW%occ(io) > 1.d-8)  THEN 
            DO i=1,n
               PAW%den(i)= PAW%den(i)+PAW%occ(io)*(PAW%phi(i,io)**2)
               PAW%tden(i)= PAW%tden(i)+PAW%occ(io)*(PAW%tphi(i,io)**2)
            ENDDO
         ENDIF
      ENDDO

      ALLOCATE(PS%den(n),PS%rv(n),stat=i)
      IF (i /= 0) THEN 
         WRITE(6,*) 'Error in FindVlocfromVeff -- ',n,i
         STOP
      ENDIF

      DO i=1,n
         PS%den(i)=FC%coreden(i)+PAW%den(i)-PAW%tden(i)
      ENDDO

      qeff=-nz+overint(n,h,PS%den)
      WRITE(6,*) 'qeff = ',qeff
      tq=overint(n,h,PAW%tden)
      WRITE(6,*) 'tq  = ',tq
      q00=-nz+FC%zcore+FC%zvale-tq
      WRITE(6,*) 'q00 = ',q00
      !  generate effective potential
      !
      PS%den=PAW%tden; PS%q=tq
      CALL potential(Grid,PS,ecoul,etxc,eexc)
      !
      rat=0
      DO i=1,n
         PAW%vloc(i)=(PAW%rveff(i)-PS%rv(i)-q00*PAW%hatpot(i))/r(i)
         IF (i>=irc) rat=rat+ABS(PAW%vloc(i))
      ENDDO
      PAW%vloc(irc:n)=0
      WRITE(6,*) 'Error in vloc -- ', rat
      DEALLOCATE(PS%rv,PS%den)
    END SUBROUTINE FindVlocfromVeff

    !*************************************************************************
    !  program to generate projector functions for Blochl's paw formalism
    !    starting with smooth functions
    !    for every basis function phi, choose smooth function with
    !     form for r<rc: (r**(l+1))*sum(n)*(cn*(r**(2*n)))
    !       where, rc==r(iiirc)
    !  phi(i,io), eval(io), tocc(io)  original basis function, energy, occupancy
    !  tphi(i,io) smooth basis function corresponding to phi(i,io)
    !  tp(i,io)=constant*normalized Harmonic Oscillator function
    !  tp == 0 for r>r(iiirc)
    !  functions defined to be identically zero for r>r(iiirc))
    !*************************************************************************
    SUBROUTINE formprojectors(Grid,Pot,PAW)
      TYPE(GridInfo),  INTENT(IN) :: Grid
      TYPE(PotentialInfo), INTENT(IN) :: Pot
      TYPE(PseudoInfo), INTENT(INOUT) :: PAW

      INTEGER :: nbase,lmax,l,io,wantednodes,nb,n,icount
      INTEGER ::  istart,ifinish,ibase,jbase
      REAL(8) :: h,xx,yy,choice
      n=Grid%n
      h=Grid%h
      lmax=PAW%lmax
      nbase=PAW%nbase
      !
      !   form projector functions for each l
      !
      DO l=0,lmax
         icount=0
         DO io=1,nbase
            IF (PAW%l(io)==l) THEN
               IF  (icount==0) istart=io
               IF  (icount >= 0) ifinish=io
               icount=icount+1
               wantednodes=icount-1
               ! form unorthonormalized projector functions tp
               WRITE(6,*) '******* projector for l = ',l
               CALL  bsolv(Grid,Pot,PAW,io,wantednodes)
            ENDIF
         ENDDO
         !write(6,*) 'orthnormalization'
         !write(6,*) 'start orthogonalization',istart,ifinish
         DO ibase=istart,ifinish
            DO jbase=istart,ibase
               IF (jbase.LT.ibase) THEN
                  xx=overlap(n,h,PAW%tp(:,jbase),PAW%tphi(:,ibase))
                  yy=overlap(n,h,PAW%tphi(:,jbase),PAW%tp(:,ibase))
                  !write(6,*) 'before',jbase,ibase,xx,yy
                  PAW%phi(1:n,ibase)=PAW%phi(1:n,ibase)-PAW%phi(1:n,jbase)*xx
                  PAW%tphi(1:n,ibase)=PAW%tphi(1:n,ibase)-PAW%tphi(1:n,jbase)*xx
                  PAW%tp(1:n,ibase)=PAW%tp(1:n,ibase)-PAW%tp(1:n,jbase)*yy
               ELSE IF (jbase.EQ.ibase) THEN
                  xx=overlap(n,h,PAW%tp(:,jbase),PAW%tphi(:,ibase))
                  yy=overlap(n,h,PAW%tphi(:,jbase),PAW%tp(:,ibase))
                  !write(6,*) 'before',jbase,ibase,xx,yy
                  choice=1.d0/SQRT(ABS(xx))
                  PAW%tp(1:n,ibase)=PAW%tp(1:n,ibase)*DSIGN(choice,xx)
                  PAW%tphi(1:n,ibase)=PAW%tphi(1:n,ibase)*choice
                  PAW%phi(1:n,ibase)=PAW%phi(1:n,ibase)*choice
               ENDIF
            ENDDO
         ENDDO

      ENDDO
    END SUBROUTINE formprojectors
    !*****************************************************************************
    !  on input tphi=phi
    !  on output tphi recalculated for r<nrc*h
    !  l = angular momentum quantum number
    !  nodes = the number of desired nodes in tp and tphi
    !  shape function
    !*****************************************************************************
    SUBROUTINE bsolv(Grid,Pot,PAW,io,wantednodes)
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(PotentialInfo), INTENT(IN) :: Pot
      TYPE(PseudoInfo), INTENT(INOUT) :: PAW
      INTEGER, INTENT(IN) :: io,wantednodes

      REAL(8), ALLOCATABLE :: a(:),b(:),c(:),f(:),chi(:)
      REAL(8), ALLOCATABLE :: tphi(:),tp(:)
      REAL(8), POINTER :: r(:),rv(:)
      REAL(8) :: h,en,v0,ol,h2,eh2,veh2,xnorm
      INTEGER ::  l,n,num,nrc,i,j,k,ii,jj,iter,irc,nodes,ok
      REAL(8) :: cpmin,cpmax
      REAL(8) :: cp,cph2,del,deriv,tderiv
      REAL(8), PARAMETER :: small=1.d-10,step=1.d0
      INTEGER, PARAMETER :: mxiter=1000

      n=Grid%n
      h=Grid%h
      r=>Grid%r
      v0=Pot%v0
      rv=>Pot%rv
      irc=PAW%irc
      ALLOCATE(a(n),b(n),c(n),f(n),chi(n),tphi(n),tp(n),stat=ok)
      IF (ok /= 0) THEN
         WRITE(6,*) 'Error in bsolv allocation ', n,ok
         STOP  
      ENDIF
      tphi=PAW%phi(:,io)
      en=PAW%eig(io)
      l=PAW%l(io)
      deriv=(tphi(irc+1)-tphi(irc-1))/(2*h*tphi(irc))  ! log derivative
      tderiv=0    ! initial log derive
      cp=0         ! initial projector constant
      del=1000     ! initial error
      cpmin=-100
      cpmax=100
      DO i=1,n
         f(i)=PAW%projshape(i)
      ENDDO

      WRITE(6,*) 'in bsolv -- l, en, n',l,en,wantednodes
      ol=l*(l+1)
      h2=h*h

      iter=0

      DO

         iter=iter+1

         WRITE(6,*) 'bsolv iter cp',iter,cp

         eh2=en*h2
         cph2=cp*h2
         veh2=(v0-en-cp)*h2


         DO i=1,irc
            b(i)=-2-(ol/(i*i)+h*rv(i)/i-eh2-cph2*f(i))/1.2d0
         ENDDO
         ! correction for l=1 case
         IF (l.EQ.1) b(1)=b(1)-1.d0/(6*(1+0.1d0*veh2))

         a(1)=0
         DO i=1,irc
            a(i+1)=1-(ol/(i*i)+h*rv(i)/i-eh2-cph2*f(i))/12.d0
         ENDDO

         DO i=2,irc+2
            c(i-1)=1-(ol/(i*i)+h*rv(i)/i-eh2-cph2*f(i))/12.d0
         ENDDO

         chi(1)=h**(l+1)
         chi(2)=-b(1)*chi(1)/c(1)
         nodes=0

         DO i=2,irc
            chi(i+1)=-(b(i)*chi(i)+a(i)*chi(i-1))/c(i)
            IF(chi(i+1)*chi(i).LE.0.d0) nodes=nodes+1
         ENDDO

         WRITE(6,'("iter nodes cp cpmin cpmax",2i5,1p3e15.7)')& 
              iter,nodes,cp,cpmin,cpmax
         IF (nodes.EQ.wantednodes) THEN
            tderiv=(chi(irc+1)-chi(irc-1))/(2*h*chi(irc))  ! log derivative
            tp=0
            tp(1:irc)=chi(1:irc)/chi(irc)
            chi(1:irc)=f(1:irc)*(tp(1:irc))**2
            xnorm=overint(irc,h,chi(1:irc))
            del=(tderiv-deriv)/xnorm
            WRITE(6,*) 'iter nodes del', iter,nodes,del

            IF (ABS(del).LT.small) EXIT

            IF (iter.GE.mxiter) THEN
               WRITE(6,*)' terminating projector',iter
               STOP
            ENDIF

            IF (ABS(del).GT.step) del=DSIGN(step,del)
            cp=cp+del
            IF (cp.GT.cpmax) cp=cpmax-ranx()*step
            IF (cp.LT.cpmin) cp=cpmin+ranx()*step

         ELSE IF(nodes.GT.wantednodes) THEN
            cpmax=cp
            cp=cpmax-ranx()*step
         ELSE IF(nodes.LT.wantednodes) THEN
            cpmin=cp
            cp=cpmin+ranx()*step 
         ENDIF

      ENDDO

      tphi(1:irc)=tphi(irc)*tp(1:irc)
      tphi(irc+1:n)=PAW%phi(irc+1:n,io)
      tp(1:irc)=f(1:irc)*tphi(1:irc)
      chi(1:irc)=tphi(1:irc)*tp(1:irc)
      xnorm=overint(irc,h,chi(1:irc))
      WRITE(6,*) 'normalization for projector l,n=',l,nodes,xnorm
      tp(1:irc)=tp(1:irc)/xnorm
      tp(irc+1:n)=0

      PAW%tphi(:,io)=tphi
      PAW%tp(:,io)=tp
      PAW%ck(io)=cp

      WRITE(6,*) 'completed bsolv',io,cp

      DEALLOCATE(a,b,c,f,chi,tphi,tp)
    END SUBROUTINE bsolv

    SUBROUTINE trunk(n,h,f,rstart,rend)
      INTEGER , INTENT(IN) :: n
      REAL(8),   INTENT(IN) :: h
      REAL(8),  INTENT(INOUT) :: f(1:n)
      REAL(8),  INTENT(IN) :: rstart,rend

      INTEGER :: i
      REAL(8) :: r,pi,delta,arg

      pi=ACOS(-1.d0)
      delta=rend-rstart
      DO i=1,n
         r=i*h
         IF (r>rstart .AND. r<=rend) THEN
            arg=pi*(r-rstart)/delta
            f(i)=f(i)*(SIN(arg)/arg)**2
         ENDIF
         IF (r>rend) f(i)=0
      ENDDO
    END SUBROUTINE trunk

    !***********************************************************************
    !  program to calculate Fourier transform of paw product tp*tphi
    !   in order to get an idea of the convergence
    !   and output them to files tprod.l
    !***********************************************************************
    SUBROUTINE ftprod(Grid,PAW)
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(PseudoInfo), INTENT(INOUT) :: PAW

      INTEGER, PARAMETER :: nq=200
      REAL(8), PARAMETER :: qmax=15.d0

      REAL(8), ALLOCATABLE :: q(:),dum(:),dum1(:)
      REAL(8), POINTER :: r(:)
      REAL(8) :: h,dq,tphij
      INTEGER :: i,ib,l,iq,n,irc
      CHARACTER(4) flnm
      !
      WRITE(6,*) 'calculating Fourier transforms of tp*tphi products  ',&
           'For bound states only '

      n=Grid%n
      h=Grid%h
      r=>Grid%r
      irc=PAW%irc
      ALLOCATE(q(nq),dum(n),dum1(n),stat=i)
      IF (i/=0) THEN
         WRITE(6,*) 'Error in allocating space in ftprod',n,nq,i
         STOP
      ENDIF

      dq=qmax/nq
      DO ib=1,PAW%nbase
         IF (PAW%eig(ib)< 0.d0) THEN
            l=PAW%l(ib)
            CALL mkname(ib,flnm)
            OPEN(55,file='tprod.'//TRIM(flnm),form='formatted')
            DO iq=1,nq
               q(iq)=iq*dq
               DO i=1,n
                  dum(i)=q(iq)*r(i)
               ENDDO
               CALL sphbes(l,n,dum)
               DO i=1,n
                  dum1(i)=dum(i)*r(i)*PAW%tphi(i,ib)
                  IF (i.LE.irc) dum(i)=dum(i)*r(i)*PAW%tp(i,ib)
               ENDDO
               tphij=overint(n,h,dum1(1:n))*overint(irc,h,dum(1:irc))

               WRITE(55,'(1p2e16.7)') q(iq),tphij
            ENDDO
            CLOSE(55)
         ENDIF  !(e<=0)
      ENDDO  ! ib
      DEALLOCATE(q,dum,dum1)
    END SUBROUTINE ftprod
    !***********************************************************************
    !  program to calculate Fourier transform of hatpot functions
    !   in order to get an idea of the convergence
    !   and output them to files hatpot.l
    !***********************************************************************
    SUBROUTINE fthatpot(Grid,PAW)
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(PseudoInfo), INTENT(INOUT) :: PAW

      INTEGER, PARAMETER :: nq=200
      REAL(8), PARAMETER :: qmax=15.d0

      REAL(8), ALLOCATABLE :: q(:),dum(:),dum1(:),dum2(:),dum3(:),arg(:)
      REAL(8), POINTER :: r(:)
      REAL(8) :: h,dq,fthatp,fthatd,fthattd
      INTEGER :: i,ib,l,iq,n,irc,ll
      CHARACTER(4) flnm
      !
      WRITE(6,*) 'calculating Fourier transforms of hatpot for  each l '

      n=Grid%n
      h=Grid%h
      r=>Grid%r
      irc=PAW%irc
      ALLOCATE(q(nq),dum(n),dum1(n),dum2(n),dum3(n),arg(n),stat=i)
      IF (i/=0) THEN
         WRITE(6,*) 'Error in allocating space in fthatpot',n,nq,i
         STOP
      ENDIF

      dq=qmax/nq
      ll=2*MAXVAL(PAW%l(:))
      DO l=0,ll
            CALL mkname(l,flnm)
            OPEN(55,file='fthatpot.'//TRIM(flnm),form='formatted')
            call hatL(Grid,PAW,l,dum2)
            DO iq=1,nq
               q(iq)=iq*dq
               DO i=1,n
                  dum(i)=q(iq)*r(i)
               ENDDO
               CALL sphbes(l,n,dum)
               DO i=1,n
                  dum3(i)=dum(i)*dum2(i)
               ENDDO
               fthatd=overint(n,h,dum3(1:n))
               fthatp=8*PI*fthatd/q(iq)**2

               fthattd=0
               if(l==0) then
                 arg(1:n)=dum(1:n)*PAW%tden(1:n)
                 fthattd=overint(n,h,arg)
               endif

               WRITE(55,'(1p5e16.7)') q(iq),fthatp,fthatd,fthatp*fthatd,fthatp*fthattd
            ENDDO
            CLOSE(55)
      ENDDO  ! l 
      DEALLOCATE(q,dum,dum1,dum2,dum3,arg)
    END SUBROUTINE fthatpot

    !***********************************************************************
    !  program to calculate Fourier transform of tphi*q^2
    !   in order to get an idea of the convergence of kinetic energy
    !   and output them to files ftkin.l
    !***********************************************************************
    SUBROUTINE ftkin(Grid,PAW)
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(PseudoInfo), INTENT(INOUT) :: PAW

      INTEGER, PARAMETER :: nq=200
      REAL(8), PARAMETER :: qmax=15.d0

      REAL(8), ALLOCATABLE :: q(:),dum(:),dum1(:)
      REAL(8), POINTER :: r(:)
      REAL(8) :: h,dq,kin
      INTEGER :: i,ib,l,iq,n,irc
      CHARACTER(4) flnm
      !
      WRITE(6,*) 'calculating Fourier transforms of tphi  ',&
           'For bound states only '

      n=Grid%n
      h=Grid%h
      r=>Grid%r
      ALLOCATE(q(nq),dum(n),dum1(n),stat=i)
      IF (i/=0) THEN
         WRITE(6,*) 'Error in allocating space in ftkin',n,nq,i
         STOP
      ENDIF

      dq=qmax/nq
      DO ib=1,PAW%nbase
         IF (PAW%eig(ib)<=0.d0) THEN
            l=PAW%l(ib)
            CALL mkname(ib,flnm)
            OPEN(55,file='ftkin.'//TRIM(flnm),form='formatted')
            DO iq=1,nq
               q(iq)=iq*dq
               DO i=1,n
                  dum(i)=q(iq)*r(i)
               ENDDO
               CALL sphbes(l,n,dum)
               DO i=1,n
                  dum1(i)=dum(i)*r(i)*PAW%tphi(i,ib)
               ENDDO
               kin=overint(n,h,dum1(1:n))*(q(iq))
               kin=kin*kin

               WRITE(55,'(1p2e16.7)') q(iq),kin
            ENDDO
            CLOSE(55)
         ENDIF  !(e<=0)
      ENDDO  ! ib
      DEALLOCATE(q,dum,dum1)
    END SUBROUTINE ftkin
    
    !***********************************************************************
    !  program to calculate Fourier transform of vloc and  tden
    !   in order to get an idea of their convergence 
    !   and output them to files ftvloc
    !***********************************************************************
    SUBROUTINE ftvloc(Grid,PAW)
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(PseudoInfo), INTENT(INOUT) :: PAW

      INTEGER, PARAMETER :: nq=200
      REAL(8), PARAMETER :: qmax=15.d0

      REAL(8), ALLOCATABLE :: q(:),dum(:),dum1(:)
      REAL(8), POINTER :: r(:)
      REAL(8) :: h,dq,vloc,tden
      INTEGER :: i,ib,l,iq,n,irc
      CHARACTER(4) flnm
      !
      WRITE(6,*) 'calculating Fourier transforms of vloc and tden'

      n=Grid%n
      h=Grid%h
      r=>Grid%r
      ALLOCATE(q(nq),dum(n),dum1(n),stat=i)
      IF (i/=0) THEN
         WRITE(6,*) 'Error in allocating space in ftvloc',n,nq,i
         STOP
      ENDIF

      OPEN(55,file='ftvloc',form='formatted')
      dq=qmax/nq
      l=0
            DO iq=1,nq
               q(iq)=iq*dq
               DO i=1,n
                  dum(i)=q(iq)*r(i)
               ENDDO
               CALL sphbes(l,n,dum)
               DO i=1,n
                  dum1(i)=dum(i)*PAW%vloc(i)*(r(i)**2)
                  dum(i)=dum(i)*PAW%tden(i)
               ENDDO
               vloc=overint(n,h,dum1(1:n))
               tden=overint(n,h,dum(1:n))

               WRITE(55,'(1p4e16.7)') q(iq),vloc,tden,vloc*tden
            ENDDO
      CLOSE(55)

      DEALLOCATE(q,dum,dum1)
    END SUBROUTINE ftvloc

    !************************************************************************
    !  pgm to integrate outward the radial schroedinger homogeneous equation
    !    at energy 'energy' and at angular momentum l
    !
    !    with potential rv/r, given in uniform mesh of n points
    !   r=i*h, i=1,...n-1 ;assuming p(r)=C*r**(l+1)*polynomial(r) for r==0;
    !                               p((n+1)*h)=0
    !  nz=nuclear charge
    !
    !  uses Noumerov algorithm
    !
    !  For l=0,1 corrections are needed to approximate wfn(r=0)
    !     These depend upon:
    !         e0 (current guess of energy eigenvalue)
    !         l,z
    !         v(0) == v0 electronic potential at r=0
    !         v'(0) == v0p derivative of electronic potential at r=0
    !
    ! also returns node == number of nodes for calculated state
    !************************************************************************
    SUBROUTINE outwardh(Grid,Pot,nr,l,energy,wfn,node)
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(PotentialInfo), INTENT(IN) :: Pot
      INTEGER, INTENT(IN) :: nr,l
      REAL(8), INTENT(IN) :: energy
      REAL(8), INTENT(INOUT) :: wfn(:)
      INTEGER, INTENT(INOUT) :: node

      REAL(8), ALLOCATABLE :: a(:),b(:)
      INTEGER :: n,i,j
      REAL(8) :: h,scale,b0p0,c1,c2,c3,ssl,ssd,angm,eh2,fact

      n=Grid%n; h=Grid%h  

      ALLOCATE(a(nr),b(nr),stat=i)
      IF (i/=0) THEN
         WRITE(6,*) 'Error in outwardh ', nr,i
         STOP 
      ENDIF

      angm=l*(l+1)
      eh2=energy*h*h
      DO i=1,nr
         a(i)=angm/(i*i)+h*Pot%rv(i)/i
      ENDDO
      DO  i=1,nr
         b(i)=-1.2d0+0.1d0*a(i)
      ENDDO
      ssl=0.1d0
      ssd=1.d0
      DO  i=1,nr
         a(i)=2.4d0+a(i)
      ENDDO
      !
      !  start outward integration
      !    correct behavior near r=0
      b0p0=0
      IF (l <= 1) THEN
         c1=-Pot%nz/(l+1.d0)
         c2=((Pot%v0-energy)-2*Pot%nz*c1)/(4*l+6.d0)
         c3=(Pot%v0p+(Pot%v0-energy)*c1-2*Pot%nz*c2)/(6*l+12.d0)
         fact=1.d0+h*(c1+h*(c2+h*c3))
         IF (l.EQ.0) THEN
            b0p0=-0.2d0*h*Pot%nz/fact
         ENDIF
         IF (l.EQ.1) THEN
            b0p0=0.2d0/fact
         ENDIF
      ENDIF
      node=0
      wfn(1)=h**(l+1)
      wfn(2)=(a(1)+b0p0-eh2*ssd)*wfn(1)/(eh2*ssl-b(2))
      DO j=3,nr
         wfn(j)=((a(j-1)-eh2*ssd)*wfn(j-1)+                               &
              (b(j-2)-eh2*ssl)*wfn(j-2))/(eh2*ssl-b(j))
         IF (wfn(j)*wfn(j-1).LE.0.d0) node=node+1
      ENDDO
      !
      ! normalize to unity within integration range
      !
      scale=1.d0/overlap(nr,h,wfn,wfn)
      scale=SQRT(scale)
      wfn(1:nr)=wfn(1:nr)*scale 
      DEALLOCATE(a,b)
    END SUBROUTINE outwardh

    !*************************************************************************
    !  pgm to integrate outward the radial schroedinger inhomogeneous equation
    !    at energy 'energy' and at angular momentum l
    !
    !    with potential smooth rveff/r, given in uniform mesh of n points
    !        proj/r == inhomogeneous function
    !     It is assumed that for r ~0, proj~~(r**(l+1))*(c0+r**2*c2+...)
    !
    !   r=i*h, i=1,...n-1 ;assuming p(r)=C*r**(l+3)*polynomial(r) for r==0;
    !
    !  uses Noumerov algorithm
    !
    !*************************************************************************
    SUBROUTINE outwardi(Grid,rveff,proj,nr,l,energy,wfn)
      TYPE(GridInfo), INTENT(IN) :: Grid
      REAL(8), INTENT(IN) :: rveff(:),proj(:)
      INTEGER, INTENT(IN) :: nr,l
      REAL(8), INTENT(IN) :: energy
      REAL(8), INTENT(INOUT) :: wfn(:)

      REAL(8), ALLOCATABLE :: a(:),b(:)
      INTEGER :: n,i,j
      REAL(8) :: h,scale,b0p0,c1,c2,c3,ssl,ssd,angm,eh2,fact
      REAL(8) :: x1,x2,x3,fac,proj0,ww

      n=Grid%n; h=Grid%h

      ALLOCATE(a(nr),b(nr),stat=i)
      IF (i/=0) THEN
         WRITE(6,*) 'Error in outwardi ', nr,i
         STOP
      ENDIF

      angm=l*(l+1)
      eh2=energy*h*h
      ! r~~0 behavior of proj :
      x3=3**(l+1)
      x2=2**(l+1)
      x1=1
      fac=1.d0/(h**(l+1))
      proj0=(proj(3)/x3+3*(proj(1)-proj(2)/x2))*fac
      !
      DO  i=1,nr
         a(i)=angm/(i*i)+h*rveff(i)/i
      ENDDO
      DO  i=1,nr
         b(i)=-1.2d0+0.1d0*a(i)
      ENDDO
      ssl=0.1d0
      ssd=1.d0
      DO  i=1,nr
         a(i)=2.4d0+a(i)
      ENDDO
      !
      !  start outward integration
      !
      ww=0.1d0*h*h
      c2=-1*proj0/(4*l+6.d0)
      wfn(1)=h**(l+3)*c2
      wfn(2)=((a(1)-eh2*ssd)*wfn(1)              &
           -ww*(proj(2)+10*proj(1)))/(eh2*ssl-b(2))
      DO j=3,nr
         wfn(j)=((a(j-1)-eh2*ssd)*wfn(j-1)       &
              -ww*(proj(j)+proj(j-2)+10*proj(j-1))      &
              +(b(j-2)-eh2*ssl)*wfn(j-2))/(eh2*ssl-b(j))
      ENDDO
      !
      DEALLOCATE(a,b)
    END SUBROUTINE outwardi

    !*******************************************************************
    !  function to calculated <wfn|O|wfn> for smooth paw wavefunction
    !*******************************************************************
    FUNCTION sepnorm(Grid,PAW,nr,l,wfn)
      REAL(8) :: sepnorm
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(PseudoInfo), INTENT(IN) :: PAW
      INTEGER, INTENT(IN) :: nr,l
      REAL(8), INTENT(IN) :: wfn(:)

      INTEGER :: n,ib,ic,nbase
      REAL(8) :: h
      REAL(8), ALLOCATABLE :: b(:)

      n=Grid%n; h=Grid%h; nbase=PAW%nbase
      ALLOCATE(b(nbase),stat=ib)
      IF (ib/=0) THEN
         WRITE(6,*) 'Error in sepnorm allocation', nbase,ib
         STOP
      ENDIF

      sepnorm=overlap(nr,h,wfn,wfn)
      b=0
      DO ib=1,nbase
         IF (l==PAW%l(ib)) b(ib)=overlap(nr,h,wfn,PAW%tp(:,ib))
      ENDDO
      DO ib=1,nbase
         DO ic=1,nbase
            sepnorm=sepnorm+b(ib)*PAW%oij(ib,ic)*b(ic)
         ENDDO
      ENDDO

      DEALLOCATE(b)   
    END FUNCTION sepnorm

    !***************************************************************************
    !  pgm to solve separable radial schroedinger equation
    !    at energy 'energy' and at angular momentum l
    !
    !    with smooth potential rveff/r, given in uniform mesh of n points
    !   r=i*h, i=1,...n-1 ;assuming p(r)=C*r**(l+1)*polynomial(r) for r==0;
    !                               p((n+1)*h)=0
    !
    !  uses Noumerov algorithm
    !
    !  For l=0,1 corrections are needed to approximate wfn(r=0)
    !     These depend upon:
    !         e0 (current guess of energy eigenvalue)
    !         l,nz==0
    !         v(0) == v0 electronic potential at r=0
    !         v'(0) == v0p derivative of electronic potential at r=0
    !
    ! also returns node == number of nodes for calculated state
    !
    !  proj == projector functions
    !  hij and qij == hamiltonianian and overlap matrix elements
    !***************************************************************************
    SUBROUTINE unboundsep(Grid,Pot,PAW,nr,l,energy,wfn,node)
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(PotentialInfo), INTENT(IN) :: Pot
      TYPE(PseudoInfo), INTENT(IN) :: PAW
      INTEGER, INTENT(IN) :: l,nr
      REAL(8), INTENT(IN) :: energy
      REAL(8), INTENT(INOUT) :: wfn(:)
      INTEGER, INTENT(INOUT) :: node

      INTEGER :: n,ia,ib,ic,nbase,icount,jcount,lcount
      REAL(8) :: summ,h,scale
      REAL(8), ALLOCATABLE :: y(:,:),b(:),a(:,:)

      n=Grid%n; h=Grid%h; nbase=PAW%nbase
      !
      CALL outwardh(Grid,Pot,nr,l,energy,wfn,node)
      !
      ALLOCATE(y(nr,nbase),b(nbase),stat=ib)
      IF (ib/=0) THEN
         WRITE(6,*) 'Error in unboundsep allocation',  nr,nbase,ib
         STOP 
      ENDIF

      lcount=0
      DO ib=1,nbase
         IF (l==PAW%l(ib)) THEN
            lcount=lcount+1
            CALL outwardi(Grid,PAW%rveff,PAW%tp(:,ib),nr,l,energy,y(:,lcount))
         ENDIF
      ENDDO
      !
      IF(lcount>0) THEN
         ALLOCATE(a(lcount,lcount),stat=ib)
         IF (ib/=0) THEN
            WRITE(6,*) 'Error in unboundsep allocation',  lcount,ib
            STOP 
         ENDIF

         icount=0
         DO ib=1,nbase
            summ=0.d0
            IF (l==PAW%l(ib)) THEN
               icount=icount+1
               DO ic=1,nbase
                  IF (l==PAW%l(ic)) THEN
                     summ=summ+(PAW%dij(ib,ic)-energy*PAW%oij(ib,ic))*&
                          overlap(nr,h,PAW%tp(:,ic),wfn)
                  ENDIF
               ENDDO
               b(icount)=-summ
            ENDIF
         ENDDO
         !
         icount=0
         DO ia=1,nbase
            IF (l==PAW%l(ia)) THEN
               icount=icount+1
               jcount=0
               DO ib=1,nbase
                  IF (l==PAW%l(ib)) THEN
                     jcount=jcount+1
                     summ=0.d0
                     IF (ia.EQ.ib) summ=1.d0
                     DO ic=1,nbase
                        IF (l==PAW%l(ic)) THEN
                           summ=summ+(PAW%dij(ia,ic)-energy*PAW%oij(ia,ic))*&
                                overlap(nr,h,PAW%tp(:,ic),y(1,jcount))
                        ENDIF
                     ENDDO
                     a(icount,jcount)=summ
                  ENDIF
               ENDDO
            ENDIF
         ENDDO
         !
         CALL linsol(a,b,lcount)

         icount=0
         DO ib=1,nbase
            IF(l==PAw%l(ib)) THEN
               icount=icount+1
               wfn(1:nr)=wfn(1:nr)+b(icount)*y(1:nr,icount)
            ENDIF
         ENDDO
         DEALLOCATE(a)
      ENDIF
      !
      ! normalize to unity within integration range
      !
      scale=1.d0/sepnorm(Grid,PAW,nr,l,wfn)
      IF (scale.LE.0.d0) THEN
         WRITE(6,*) 'warning -- negative norm for l=',l
         scale=-scale
         IF (scale.EQ.0.d0) scale=1.d0
      ENDIF
      scale=DSIGN(SQRT(scale),wfn(nr-2))
      wfn(1:nr)=wfn(1:nr)*scale 
      DEALLOCATE(b,y)   
    END SUBROUTINE unboundsep

    !*********************************************************
    !  program to to transform smooth wavefunction to all-electron
    !     wavefunction within Blochl's paw formalism
    !   otp == projector function
    !   odphi == ophi - otphi
    !*********************************************************
    SUBROUTINE PStoAE(Grid,PAW,nr,l,tpsi,psi)
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(PseudoInfo), INTENT(IN) :: PAW
      INTEGER, INTENT(IN) :: nr,l
      REAL(8), INTENT(IN) :: tpsi(:)
      REAL(8), INTENT(INOUT) :: psi(:)

      INTEGER :: n,ib,nbase
      REAL(8) :: h,scale

      n=Grid%n; h=Grid%h; nbase=PAW%nbase

      psi(1:nr)=tpsi(1:nr)
      DO ib=1,nbase
         IF (l==PAW%l(ib)) psi(1:nr)=psi(1:nr)+&
              (PAW%phi(1:nr,ib)-PAW%tphi(1:nr,ib))*&
              overlap(nr,h,PAW%tp(:,ib),tpsi)
      ENDDO
    END SUBROUTINE PStoAE

    !************************************************************************
    !  program to calculate logerivatives of paw wavefunctions
    !   and to compare them with all electron wavefunctions
    !  optionally, wavefunctions are written to a file
    !************************************************************************
    SUBROUTINE logderiv(Grid,Pot,PAW)
      TYPE(GridInfo), INTENT(IN) :: Grid
      TYPE(PotentialInfo), INTENT(IN) :: Pot
      TYPE(PseudoInfo), INTENT(INOUT) :: PAW

      TYPE(PotentialInfo) :: PS
      INTEGER :: n,l,ie,nbase,ib,ic,nr,i,nodes,mbase,irc
      INTEGER, PARAMETER :: ne=200,ke=100
      REAL(8), PARAMETER :: e0=-5.d0,de=0.05d0
      REAL(8) :: h,x,dwdr,dcwdr,scale,energy
      REAL(8), ALLOCATABLE :: psi(:),tpsi(:),ttpsi(:)
      CHARACTER(4)  :: flnm

      n=Grid%n; h=Grid%h; nbase=PAW%nbase; irc=PAW%irc;  nr=irc+10

      ALLOCATE(psi(nr),tpsi(nr),ttpsi(nr),PS%rv(n),stat=ie)
      IF (ie/=0) THEN
         WRITE(6,*) 'Error in logderiv allocation',n,ie
         stop
      ENDIF

      !  load PAW%dij and PAW%oij

      PAW%dij=0; PAW%oij=0

      DO ib=1,nbase
         DO ic=ib,nbase
            IF (PAW%l(ib)==PAW%l(ic)) THEN
               CALL dqij(Grid,PAW,ib,ic,PAW%oij(ib,ic))
               CALL dtij(Grid,PAW,ib,ic,PAW%dij(ib,ic))
               CALL avij(Grid,Pot,PAW,ib,ic,x)
               PAW%dij(ib,ic)=PAW%dij(ib,ic)+x
               IF  (ic>ib) THEN
                  PAW%oij(ic,ib)=PAW%oij(ib,ic)
                  PAW%dij(ic,ib)=PAW%dij(ib,ic)
               ENDIF
            ENDIF
         ENDDO
      ENDDO

      ! load  PS
      PS%rv=PAW%rveff ; PS%nz=0
      psi(1:15)=PS%rv(1:15)/Grid%r(1:15)
      CALL nderiv(h,psi,tpsi,15,i)
      PS%v0=psi(3)+3*(psi(1)-psi(2))  
      PS%v0p=tpsi(3)+3*(tpsi(1)-tpsi(2))
      !
      !   calculate logderivatives at irc
      !
      WRITE(6,*) 'calculating log derivatives at irc',Grid%r(irc)
      !
      mbase=nbase
      DO l=0,PAW%lmax+1
         CALL mkname(l,flnm)
         OPEN(56,file='logderiv.'//TRIM(flnm),form='formatted')

         DO ie=1,ne
            energy=e0+de*(ie-1)
            psi=0;tpsi=0;ttpsi=0
            CALL unboundsch(Grid,Pot,nr,l,energy,psi,nodes)
            CALL unboundsep(Grid,PS,PAW,nr,l,energy,tpsi,nodes)
            CALL PStoAE(Grid,PAW,nr,l,tpsi,ttpsi)
            !
            dwdr=(psi(irc+1)-psi(irc-1))/(2*h*psi(irc))
            dcwdr=(ttpsi(irc+1)-ttpsi(irc-1))/(2*h*ttpsi(irc))

            WRITE(56,'(1p5e12.4)') energy,dwdr,dcwdr
            IF (ie.EQ.ke) THEN
               mbase=mbase+1
               CALL mkname(mbase,flnm)
               OPEN(57,file='wfn'//TRIM(flnm),form='formatted')
               WRITE(57,*) 'l=',l,'energy=',energy
               !
               ! form converted wavefunction and rescale exact wavefunction
               !
               scale=ttpsi(irc)/psi(irc)
               DO i=1,nr
                  WRITE(57, '(1p5e12.4)') Grid%r(i),tpsi(i),ttpsi(i),psi(i)*scale
               ENDDO
               CLOSE(57)
            ENDIF
         ENDDO !ie
         CLOSE(56)

      ENDDO !l

      DEALLOCATE(psi,tpsi,PS%rv)
    END SUBROUTINE logderiv

  END MODULE basis
